/* Goxel 3D voxels editor
 *
 * copyright (c) 2018 Guillaume Chereau <guillaume@noctua-software.com>
 *
 * Goxel is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.

 * Goxel is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.

 * You should have received a copy of the GNU General Public License along with
 * goxel.  If not, see <http://www.gnu.org/licenses/>.
 */

/* This file is autogenerated by tools/create_assets.py */

{.path = "data/themes/dark.ini", .size = 206, .data =
    "[theme]\n"
    "name=dark\n"
    "\n"
    "[base]\n"
    "background=#2D2D2DFF\n"
    "outline=#FF\n"
    "inner=#2B3944FF\n"
    "inner_selected=#516882FF\n"
    "text=#C8C8C8FF\n"
    "text_selected=#FF\n"
    "\n"
    "[widget]\n"
    "\n"
    "[tab]\n"
    "background=#FF\n"
    "inner=#2D2D2DFF\n"
    "\n"
    "[menu]\n"
    "inner=#5A6E6EFF\n"
    ""
},
{.path = "data/themes/original.ini", .size = 253, .data =
    "[theme]\n"
    "name=original\n"
    "\n"
    "[base]\n"
    "background=#607272FF\n"
    "outline=#4D4D4DFF\n"
    "inner=#A1A1A1FF\n"
    "inner_selected=#6666CCFF\n"
    "text=#111111FF\n"
    "text_selected=#111111FF\n"
    "\n"
    "[widget]\n"
    "\n"
    "[tab]\n"
    "background=#304242FF\n"
    "inner=#607272FF\n"
    "inner_selected=#455656FF\n"
    "\n"
    "[menu]\n"
    "inner=#5A6E6EFF\n"
    ""
},



